//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// Colour class - augmentation of Color to allow lookups by colour name
//
import java.awt.Color;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;

/**
 * Colour class augments Color to add a lookup for colours that are named
 * in properties.  Plus, it allows me to spell Colour the British way.
 */
public class Colour {

	// message and colour catalog information
	private static final ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );
	private static final ResourceBundle	rgb		=
				ResourceBundle.getBundle( "RGB" );


	/**
	 * Return a colour from a string.
	 * <p>
	 * The name given may be the logical name, looked up in the colour
	 * resources database as Colour.&lt;string&gt;, for example,
	 * "Peach_Puff", or it may be a decodable string-format 24bit
	 * representation, for example, "0xffdab9".  The corresponding
	 * colour is returned.  If the name is invalid, or the 24bit
	 * representation is invalid, the method returns Color.black, to
	 * allow the application to carry on (though perhaps in the dark?).
	 */
	public static Color find_colour( String colour_spec_or_name ) {

		// try get the colour by name from the properties database
		Color colour = lookup_colour( colour_spec_or_name );

		// if not found, try to create a new colour using the spec
		// as a decodable string
		if ( colour == null )
			colour = create_colour( colour_spec_or_name );

		// if that worked, return it
		if ( colour != null )
			return ( colour );

		// output an error message, and return black
		String[] fmtargs = { colour_spec_or_name };
		System.err.println( MessageFormat.format(
			messages.getString( "Colour.badvalue" ), fmtargs ));
		return ( Color.black );
	}

	/*
	 * Create a colour from a decodable 24bit string representation.
	 */
	public static Color create_colour( String colour_spec ) {

		// try to decode the string into a colour
		try {
			return ( Color.decode( colour_spec ) );
		}
		catch ( NumberFormatException e ) {
			String[] fmtargs = { colour_spec };
			System.err.println( MessageFormat.format(
				messages.getString(
					"Colour.badvalue" ),
				fmtargs ));
		}

		// return null on error
		return ( null );
	}

	/*
	 * Return a colour from a colour name, or null if the name is
	 * not valid.
	 */
	private static Color lookup_colour( String name ) {

		// look up the colour definition in the resource database
		String colour_spec	= null;
		try {
			colour_spec = rgb.getString( name.toLowerCase() );
		}
		catch ( MissingResourceException e ) {

			// not found, so return null
			return ( null );
		}

		// create the colour entry from the definition
		return ( create_colour( colour_spec ));
	}


	/**
	 * Class test code.
	 */
	public static void main ( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		System.out.println( find_colour( "White" ));
		System.out.println( find_colour( "Light_Gray" ));
		System.out.println( find_colour( "Peach_Puff" ));
		System.out.println( find_colour( "Black" ));
		System.out.println( find_colour( "0x000000" ));
		System.out.println( find_colour( "0xABCDEF" ));
		System.out.println( find_colour( "0xFFFFFF" ));
	}
}
