//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// ExerciseGUI class - creates a frame to run through an exercise lesson
//
import java.awt.Toolkit;
import java.awt.Color;
import java.awt.Frame;
import java.awt.Font;
import java.awt.Panel;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.CardLayout;
import java.awt.CheckboxMenuItem;
import java.awt.Button;
import java.awt.Label;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.MenuShortcut;
import java.awt.TextArea;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.FocusListener;
import java.awt.event.FocusEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.text.MessageFormat;

/**
 * A window that runs through an exercise lesson.
 */
public class ExerciseGUI extends Frame
			implements Runnable,
			KeyListener, FocusListener, ActionListener,
			WindowListener {
	private boolean inAnApplet = true;

	// message and colour catalog information
	private static final ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );
	private static final ResourceBundle	colours		=
				ResourceBundle.getBundle( "ColorResource" );

	// the graphics toolkit, for beeping on mistakes
	private static final Toolkit		toolkit		=
						Toolkit.getDefaultToolkit();

	// the number of mS between keypresses before we decide that we're
	// going to stop timing for a bit
	private static final int	MS_BEFORE_PAUSE		=
			Integer.parseInt( messages.getString(
						"ExerciseGUI.msbeforepause" ));

	// symbols used to show bad keys struck
	private static final String	BADKEY_INDICATOR =
				messages.getString( "ExerciseGUI.badkey" );
	private static final String	BADCR_INDICATOR	=
				messages.getString( "ExerciseGUI.badcr" );

	// statistics GUI we could refresh after a lesson
	private StatisticsGUI	statistics;//	= null;

	// keyboard GUI for showing what we expect, and what was typed
	private KeyboardGUI	keyboard;//	= null;

	// translation service, to map keypresses between keyboard layouts
	// if that's what we've been asked to do
	private	KBTranslator	translator;//	= null;

	// menubar GUI components
	private MenuBar			menu_bar;
	private Menu			file_menu, history_menu;
	private MenuItem		file_restart, file_pause, file_close;
	private Menu			options_menu;
	private CheckboxMenuItem	option_wrap, option_para,
					option_skipspace, option_hyphen;
	private CheckboxMenuItem	option_mistake_beep;
	private CheckboxMenuItem	option_capslock_workround;

	// GUI components
	private ImageButton	restart_button, pause_button;
	private Tooltip		restart_tooltip, pause_tooltip;
	private Label		title_label, descr_label;
	private TextArea	reference_text_area;
	private TextArea	entry_text_area;
	private TextArea	paragraph_text_area;
	// extra stuff to do the card layout
	private Panel		text_panel;
	private CardLayout	card_layout;
	private Label		modifiers_label;
	private Label		status_label;

	// lesson history
	private History		history;
	private Hashtable	history_menuitems;

	// the lesson we are working on, with all its attributes
	// expanded out for convenient access
	private ExerciseLesson	lesson;//		= null;
	private String		lesson_title;//		= null;
	private String		lesson_instructions;//	= null;
	private String		lesson_text;//		= null;

	// data to track progress through the data
	private KeyingStatistics
				keying_stats	= new KeyingStatistics();
	private KeyingRhythm	keying_rhythm	= new KeyingRhythm();
	private long		keypress_timestamp;//	= 0;
	private int		progress;//		= 0;
	private int		cursor;//		= 0;

	// statistics set we are working with, for reference
	private KBStatistics		stats_set;//	= null;

	// dialog to show the results of the exercise in
	private ExerciseGUIStats	finish_dialog;//= null;

	// workround for the fact that caps_lock doesn't in AWT 1.1
	private boolean		caps_lock;//				= false;
	private boolean		caps_lock_was_last_press_event;//	= false;
	private long		caps_lock_timestamp;//			= 0;

	// synchronization monitor, for setVisible notifications
	private Object		visible		= new Object();


	/**
	 * Initialize the whole GUI set.
	 */
	public ExerciseGUI( boolean inAnApplet, KBStatistics stats_set ) {
		this.addWindowListener( this );
		this.inAnApplet = inAnApplet;

		// create GUI components
		setTitle( messages.getString( "ExerciseGUI.title" ));
		setLayout( new BorderLayout() );
		setForeground( Colour.find_colour(
			colours.getString( "ExerciseGUI.foreground" )));
		setBackground( Colour.find_colour(
			colours.getString( "ExerciseGUI.background" )));
		setFont( Font.decode(
			colours.getString( "ExerciseGUI.font" )));

		// note the stats set for reference
		this.stats_set = stats_set;

		// create the menu bar stuff
		menu_bar = new MenuBar();
		setMenuBar( menu_bar );
		// File menu
		file_menu = new Menu(
			messages.getString( "ExerciseGUI.file" ));
		menu_bar.add( file_menu );
		file_restart = new MenuItem(
			messages.getString( "ExerciseGUI.file_restart" ),
			new MenuShortcut(
				messages.getString(
				"ExerciseGUI.file_restart_sc" ).charAt(0) ));
		file_restart.addActionListener( this );
		file_menu.add( file_restart );
		file_pause = new MenuItem(
			messages.getString( "ExerciseGUI.file_pause" ),
			new MenuShortcut(
				messages.getString(
				"ExerciseGUI.file_pause_sc" ).charAt(0) ));
		file_pause.addActionListener( this );
		file_menu.add( file_pause );
		file_close = new MenuItem(
			messages.getString( "ExerciseGUI.file_close" ),
			new MenuShortcut(
				messages.getString(
				"ExerciseGUI.file_close_sc" ).charAt(0) ));
		file_close.addActionListener( this );
		file_menu.add( file_close );
		// Option menu
		options_menu = new Menu(
			messages.getString( "ExerciseGUI.options" ));
		menu_bar.add( options_menu );
		option_wrap = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_wrap" ),
				true );
		options_menu.add( option_wrap );
		option_para = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_para" ),
				true );
		options_menu.add( option_para );
		option_skipspace = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_skipspace" ),
				true );
		options_menu.add( option_skipspace );
		option_hyphen = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_hyphen" ),
				true );
		options_menu.add( option_hyphen );
		options_menu.add( new MenuItem( "-" ));
		option_mistake_beep = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_mistake_beep" ),
				false );
		options_menu.add( option_mistake_beep );
		options_menu.add( new MenuItem( "-" ));
		option_capslock_workround = new CheckboxMenuItem(
			messages.getString( "ExerciseGUI.options_capslock" ),
				false );
		options_menu.add( option_capslock_workround );
		// History menu
		history_menu = new Menu(
			messages.getString( "ExerciseGUI.history" ));
		menu_bar.add( history_menu );
		history = new History( Integer.parseInt(
			messages.getString( "ExerciseGUI.historylength" )));
		history_menuitems = new Hashtable();


		// create the pause and restart buttons 
		Panel toolbar = new Panel();
		toolbar.setLayout( new FlowLayout( FlowLayout.LEFT ));

		restart_button = new ImageButton(
			Xpm.get_image_producer(
				messages.getString( "ExerciseGUI.restartxpm" )),
			messages.getString( "ExerciseGUI.restartxpmlabel" ));
		restart_tooltip = new Tooltip( this, restart_button,
			messages.getString( "ExerciseGUI.restarttooltip" ));
		restart_button.addActionListener( this );
		toolbar.add( restart_button );
		pause_button = new ImageButton(
			Xpm.get_image_producer(
				messages.getString( "ExerciseGUI.pausexpm" )),
			messages.getString( "ExerciseGUI.pausexpmlabel" ));
		pause_tooltip = new Tooltip( this, pause_button,
			messages.getString( "ExerciseGUI.pausetooltip" ));
		pause_button.addActionListener( this );
		toolbar.add( pause_button );

		// create the labels, and add them to the toolbar
		title_label = new Label();
		descr_label = new Label();
		Panel label_panel = new Panel();
		label_panel.setLayout( new GridLayout( 0, 1 ));
		label_panel.add( title_label );
		label_panel.add( descr_label );
		Panel top_panel = new Panel();
		top_panel.setLayout( new BorderLayout() );
		top_panel.add( "West", toolbar );
		top_panel.add( "Center", label_panel );
		add( "North", top_panel );

		// create the text displays
		text_panel = new Panel();
		card_layout = new CardLayout();
		text_panel.setLayout( card_layout );

		Panel drill_panel = new Panel();
		drill_panel.setLayout( new GridLayout( 0, 1 ));
		reference_text_area = new TextArea(
				Integer.parseInt(
			messages.getString( "ExerciseGUI.referencerows" )),
				Integer.parseInt(
			messages.getString( "ExerciseGUI.referencecolumns" )));
		reference_text_area.setForeground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.upperdrillpanel_foreground" )));
		reference_text_area.setBackground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.upperdrillpanel_background" )));
		reference_text_area.setFont( Font.decode(
				colours.getString(
				"ExerciseGUI.upperdrillpanel_font" )));
		reference_text_area.addKeyListener( this );
		reference_text_area.addFocusListener( this );
		reference_text_area.setEditable( false );
		drill_panel.add( reference_text_area );
		entry_text_area = new TextArea(
				Integer.parseInt(
			messages.getString( "ExerciseGUI.entryrows" )),
				Integer.parseInt(
			messages.getString( "ExerciseGUI.entrycolumns" )));
		entry_text_area.setForeground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.lowerdrillpanel_foreground" )));
		entry_text_area.setBackground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.lowerdrillpanel_background" )));
		entry_text_area.setFont( Font.decode(
				colours.getString(
				"ExerciseGUI.lowerdrillpanel_font" )));
		entry_text_area.addKeyListener( this );
		entry_text_area.addFocusListener( this );
		entry_text_area.setEditable( false );
		drill_panel.add( entry_text_area );

		Panel paragraph_panel = new Panel();
		paragraph_panel.setLayout( new GridLayout( 0, 1 ));
		paragraph_text_area = new TextArea(
				Integer.parseInt(
			messages.getString( "ExerciseGUI.paragraphrows" )),
				Integer.parseInt(
			messages.getString( "ExerciseGUI.paragraphcolumns" )));
		paragraph_text_area.setForeground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.paragraphpanel_foreground" )));
		paragraph_text_area.setBackground( Colour.find_colour(
			colours.getString(
				"ExerciseGUI.paragraphpanel_background" )));
		paragraph_text_area.setFont( Font.decode(
				colours.getString(
				"ExerciseGUI.paragraphpanel_font" )));
		paragraph_text_area.addKeyListener( this );
		paragraph_text_area.addFocusListener( this );
		paragraph_text_area.setEditable( false );
		paragraph_panel.add( paragraph_text_area );

		text_panel.add( "DRILL_PANEL", drill_panel );
		text_panel.add( "PARAGRAPH_PANEL", paragraph_panel );
		card_layout.show( text_panel, "DRILL_PANEL" );
		add( "Center", text_panel );

		Panel status_panel = new Panel();
		status_panel.setLayout( new GridLayout( 1, 0 ));
		modifiers_label = new Label( "", Label.LEFT );
		status_panel.add( modifiers_label );
		status_label = new Label(
			messages.getString( "ExerciseGUI.nolesson" ),
				Label.RIGHT );
		status_panel.add( status_label );
		add( "South", status_panel );

		// create the dialog for the completion of the exercise
		finish_dialog = new ExerciseGUIStats( inAnApplet, this, false );

		// finally, create a thread to maintain the status label
		new Thread( this, "ExerciseGUI_updater" ).start();
	}


	/**
	 * Thread that keeps the status line updated with what's going
	 * on in this object.
	 */
	public void run() {

		// colours for status messages
		final Color	COLOUR_NORMAL_STATUS	= getForeground();
		final Color	COLOUR_BUSY_STATUS	=
				Colour.find_colour( colours.getString(
						"ExerciseGUI.inprogress" ));

		// thread 'tick' period, in mS, and paused flash period
		final int	THREAD_TICK		= 125;
		final int	PAUSE_FLASH_PERIOD	=
				Integer.parseInt( messages.getString(
					"ExerciseGUI.pauseflashperiod" ));

		// note of the status line contents, in case we can
		// avoid the flicker of updating it when it hasn't
		// changed
		String		last_status		= "";

		// variable to detect when nothing was typed for a while,
		int		last_progress		= 0;
		long		ms_since_progress	= 0;


		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// if our window isn't visible then just do nothing
			while ( ! isVisible() ) {
				try { synchronized( visible ) {
					visible.wait();
				}}
				catch ( InterruptedException e ) {}
			}

			// wait until the next tick
			try { Thread.sleep( THREAD_TICK ); }
			catch ( InterruptedException e ) {}

			// check if there has been anything typed since
			// we last ran, and if not, how long since something
			// was typed
			if ( progress == last_progress )
				ms_since_progress += (long)THREAD_TICK;
			else
				ms_since_progress = 0L;
			last_progress = progress;

			// compute the state of the process depending on
			// what's happened since the last tick
			String	status		=
				messages.getString( "ExerciseGUI.nolesson" );
			Color	status_colour	= COLOUR_NORMAL_STATUS;

			// see if we have a lesson at all
			if ( lesson != null ) {

				// see if we are at the start of the lesson
				if ( progress == 0 || keypress_timestamp == 0 )
					status = messages.getString(
						"ExerciseGUI.ready" );

				// see if we are at the end of the lesson
				else if ( progress == lesson_text.length() )
					status = messages.getString(
						"ExerciseGUI.completed" );

				// we are midway through a lesson
				else {

					// have we paused in mid-flow?
					if ( ms_since_progress >= (long)MS_BEFORE_PAUSE ) {

						// create a blinking pause msg
						long now = System.currentTimeMillis();
						if ( now % PAUSE_FLASH_PERIOD <
							PAUSE_FLASH_PERIOD / 2 )
							status =
							messages.getString(
							"ExerciseGUI.paused" );
						else
							status = "";
					}
					else {

						// in a lesson, and not paused
						int progress_pc =
							( progress * 100 )
							/ lesson_text.length();
						Object[] fmtargs = {
						new Float( (float)progress_pc/100f )
						};
						status = MessageFormat.format(
							messages.getString(
							"ExerciseGUI.progress" ),
								fmtargs );
					}

					// colour progress messages differently
					status_colour = COLOUR_BUSY_STATUS;
				}
			}

			// compare the current status with the last status,
			// and if different update the status label
			if ( ! status.equals( last_status )) {
					status_label.setForeground(
								status_colour );
					status_label.setText( status );
					last_status = status;
			}
		}
	}


	/*
	 * Update the history menu with this lesson.
	 */
	private void update_history( ExerciseLesson lesson ) {

		// get the entry format
		String entry_format	= messages.getString(
						"ExerciseGUI.historyformat" );

		// see if this lesson is a new one to the list
		if ( ! history.remembers( lesson )) {

			// add to the lesson history
			history.remember( lesson );

			// new lesson, so update the history menu
			history_menu.removeAll();
			history_menuitems.clear();
			for ( Enumeration enum = history.relate();
					enum.hasMoreElements(); ) {

				// get the lesson title, and make a menu item
				Lesson hlesson	= (Lesson)(enum.nextElement());
				String stext	= hlesson.get_text()
							.replace( '\n', ' ' )
							.replace( '\t', ' ' );
				String[] fmtargs= { hlesson.get_title(),
					( stext.length() > 20 ) ?
						stext.substring( 0, 20 ) :
									stext };
				MenuItem menuitem = new MenuItem(
					MessageFormat.format( entry_format,
								fmtargs ));
				menuitem.addActionListener( this );

				// add this menu item to the menu and list
				history_menu.add( menuitem );
				history_menuitems.put( menuitem, hlesson );
			}
		}
	}


	/**
	 * Run the exercise from a lesson passed in.
	 */
	public void reveal( ExerciseLesson lesson ) {

		// update the lesson history menu
		update_history( lesson );

		// salt away the lesson so we can refer to it later
		this.lesson	= lesson;

		// get the lesson's title, instructions, and text
		lesson_title		= lesson.get_title();
		lesson_instructions	= lesson.get_instructions();
		lesson_text		= lesson.get_text();

		// set up the labels and text
		title_label.setText( lesson_title );
		descr_label.setText( lesson_instructions );
		if ( lesson instanceof DrillLesson ) {
			card_layout.show( text_panel, "DRILL_PANEL" );
			reference_text_area.setText( lesson_text );
			entry_text_area.setText( "" );
		}
		else { // ( lesson instanceof ParagraphLesson )
			card_layout.show( text_panel, "PARAGRAPH_PANEL" );
			paragraph_text_area.setText( lesson_text );
		}

		// start with a clean set of keying statistics
		keying_stats.reset();
		keying_rhythm.reset();

		// initialize our pointer to the start of the
		// string being used for reference
		progress	= 0;
		cursor		= 0;

		// fake an initial end of paragraph, so that any of
		// the wp modes set can take affect immediately - skip over
		// any initial indents and so on
		if ( lesson instanceof DrillLesson )
			adjust_for_wpmodes( '\n', true );
		else // ( lesson instanceof ParagraphLesson )
			adjust_for_wpmodes( '\n', false );

		// clear highlighting on the keyboard window if
		// we have been given one
		if ( keyboard != null )
			keyboard.clear_expected();

		// reset the keypress timestamp so that the first key
		// of this lesson is ignored
		keypress_timestamp = 0;

		// reveal the frame
		status_label.setText(
				messages.getString( "ExerciseGUI.ready" ));
		setVisible( true );
	}


	/*
	 * Do a selection in text areas, to highlight the next expected
	 * character the user should type
	 */
	private void show_typing_cursor() {

		// ignore if no lesson given to us yet
		if ( lesson != null ) {
			if ( lesson instanceof DrillLesson )
				reference_text_area.select(
						cursor, cursor + 1 );
			else // ( lesson instanceof ParagraphLesson )
				paragraph_text_area.select(
						cursor, cursor + 1 );
		}
	}


	/*
	 * Update the keyboard window to show the character expected next.
	 */
	private void show_expected_character() {

		// ignore if no lesson given to us yet
		if ( lesson != null && keyboard != null ) {

			// see if we are still inside a lesson
			if ( progress < lesson_text.length() )

				// highlight the expected character
				keyboard.char_expected(
						lesson_text.charAt( progress ));
		}
	}


	/**
	 * Override setVisible with something that creates the typing
	 * cursor after the component is rendered, otherwise we don't
	 * get to see it until after the first character press.
	 */
	public void setVisible( boolean is_visible ) {
		super.setVisible( is_visible );
		if ( is_visible ) {
			synchronized( visible ) { visible.notify(); }

			// if we have a current lesson select the right part
			// of it
			show_typing_cursor();

			// indicate what is next on the keyboard window
			show_expected_character();
		}
	}


	/*
	 * Handle dismiss, close, etc by hiding ourselves, and our
	 * statistics dialog popup.
	 */
	private void handle_menu_close() {
		if ( finish_dialog != null )
			finish_dialog.setVisible( false );
		setVisible( false );
	}


	/*
	 * Handle pause of the lesson by setting the last keypress timestamp
	 * to 0; this fools the rest of the class into believing that the
	 * time between keypresses is wildly excessive..
	 */
	private void handle_menu_pause() {
		keypress_timestamp = 0;
	}


	/*
	 * Handle restart of the lesson by re-calling reveal() with our
	 * saved lesson (if we have one, that is) - wonderful.
	 */
	private void handle_menu_restart() {
		if ( lesson != null )
			reveal( lesson );
	}


	/*
	 * Return the time in milliseconds between successive calls,
	 * for timing key strikes.
	 */
	private long delta_time() {

		// get the time now, and subtract from it the time we
		// last got a call
		long now_timestamp = System.currentTimeMillis();
		long delta = now_timestamp - keypress_timestamp;

		// update our note of when we were last called
		keypress_timestamp = now_timestamp;

		// now return the interval in mS
		return ( delta );
	}


	/*
	 * Update all the statistics we have to given a wanted char, and
	 * whether it was hit (true) or missed (false).
	 */
	private void update_stats( char char_struck, boolean hit_ok ) {

		// find the time since the last key was pressed
		long delta_time = delta_time();

		// if the time is something we should note, update the
		// statistics - ignore things over a certain threshold
		// to avoid skewing the statistics in the event of a
		// break in mid-exercise
		if ( delta_time > 0L && delta_time < (long)MS_BEFORE_PAUSE ) {

			// it's safe to convert delta_time into an int
			int intdelta_time = (int)delta_time;

			// first update our lesson-specific statistics
			if ( hit_ok )
				keying_stats.struck( intdelta_time );
			else
				keying_stats.missed( intdelta_time );

			// now update the global key-specific ones
			stats_set.update_statistics( char_struck, hit_ok,
								intdelta_time );

			// finally, update the rhythm monitor
			keying_rhythm.struck( intdelta_time );
		}
	}


	/*
	 * When in an exercise, we'd like to try to mimic the way a word
	 * processor works as far as we can - this is more comfortable
	 * for the majority of PC users, and not many folks will want
	 * to learn for manual typewriters (and those that do probably
	 * won't use a Java Applet to do it).
	 */
	private void adjust_for_wpmodes( char char_struck,
							boolean in_a_drill ) {

		// if the character was space, attempt to condense
		// multiple spaces into just one, by bypassing any
		// others we may find
		if ( char_struck == ' ' ) {
			if ( option_skipspace.getState() ) {
				// jump over any forthcoming spaces
				while ( progress < lesson_text.length() &&
				lesson_text.charAt( progress ) == ' ' ) {
					if ( in_a_drill )
						entry_text_area.append( " " );
					progress++; cursor++;
				}
			}
		}

		// if we were right at the end of a line and got Return,
		// try to auto-indent to a paragraph
		else if ( char_struck == '\n' ) {
			if ( option_para.getState() ) {
				// jump over both intervening spaces and nl's
				while ( progress < lesson_text.length() &&
				( lesson_text.charAt( progress ) == ' '
				  || lesson_text.charAt( progress ) == '\n' )) {
					if ( in_a_drill )
						entry_text_area.append(
							lesson_text.substring(
							progress, progress + 1 ));
					progress++; cursor++;
				}
			}
		}

		// if we have reached the end of the line but one, and
		// the line ends in a hyphen (but not two hyphens), skip it
		// and the newline too
		else if ( char_struck != '-'
			&& progress < lesson_text.length() - 1
			&& lesson_text.substring( progress, progress + 2 )
				.equals( "-\n" )) {
			if ( option_hyphen.getState() ) {
				if ( in_a_drill )
					entry_text_area.append( "-\n" );
				progress++; cursor++;
				progress++; cursor++;
			}
		}
	}


	/*
	 * Handle a key press in the frame.
	 */
	private void handle_key_press( char char_struck ) {

		// ignore if no lesson given to us yet
		if ( lesson != null ) {

			// pass the char to the appropriate handling routine
			if ( lesson instanceof DrillLesson )
				handle_key_press_drill( char_struck );
			else // ( lesson instanceof DrillLesson )
				handle_key_press_paragraph( char_struck );

			// now, if the lesson end was reached update the
			// results dialog with the timings
			if ( progress == lesson_text.length() ) {
				finish_dialog.reveal( lesson, keying_stats,
						keying_rhythm );

				// refresh the statistics GUI window if we have
				// been given one
				if ( statistics != null )
					statistics.refresh();

				// clear highlighting on the keyboard window if
				// we have been given one
				if ( keyboard != null )
					keyboard.clear_expected();
			}
		}
	}
	private void handle_key_press_drill( char char_struck ) {

		// ignore if we are at the end of the reference string
		if ( progress < lesson_text.length() ) {

			// compare the key pressed to the one expected
			char char_wanted = lesson_text.charAt( progress );
			if ( char_struck == char_wanted ||
				( option_wrap.getState() &&
				  char_wanted == '\n' && char_struck == ' ' )) {

				// update the timings and statistics
				update_stats( char_struck, true );

				// update the input text area
				entry_text_area.append(
					lesson_text.substring(
					progress, progress + 1 ));

				// advance the cursors
				progress++; cursor++;

				// indicate the success on the keyboard window
				if ( keyboard != null )
					keyboard.char_success( char_struck );

				// if wp modes set, there are some adjustments
				// to the text display that are useful here
				adjust_for_wpmodes( char_struck, true );
			}
			else {
				// wrong key - handle according to type
				// -- just ignore deletes in this case
				if ( char_struck != '\b' ) {

					// update the timings and statistics
					update_stats( char_wanted, false );

					// beep if required on wrong keys
					if ( option_mistake_beep.getState() )
						toolkit.beep();

					// update the input text area
					if ( char_wanted == '\n' ) {
						entry_text_area.append(
							BADCR_INDICATOR + "\n" );
						// advance the cursors
						progress++; cursor++;
					}
					else {
						entry_text_area.append(
							BADKEY_INDICATOR );
						// advance the cursors
						progress++; cursor++;
					}

					// indicate the mistake on the keyboard
					// window
					if ( keyboard != null )
						keyboard.char_error(
								char_struck );
				}
			}

			// highlight progress through the exercise
			show_typing_cursor();
			show_expected_character();
		}
	}
	private void handle_key_press_paragraph( char char_struck ) {

		// ignore if we are at the end of the reference string
		if ( progress < lesson_text.length() ) {

			// compare the key pressed to the one expected
			char char_wanted = lesson_text.charAt( progress );
			if ( char_struck == char_wanted ||
				( option_wrap.getState() &&
				  char_wanted == '\n' && char_struck == ' ' )) {

				// update the timings and statistics
				update_stats( char_struck, true );

				// advance the cursors
				progress++; cursor++;

				// indicate the success on the keyboard window
				if ( keyboard != null )
					keyboard.char_success( char_struck );

				// if wp modes set, there are some adjustments
				// to the text display that are useful here
				adjust_for_wpmodes( char_struck, true );
			}
			else {

				// wrong key - handle according to type
				if ( char_struck == '\b' ) {
					if ( progress > 0 && cursor > 0
					&& lesson_text.charAt( progress - 1 )
							!= '\n' ) {
						// back up one character
						progress--; cursor--;
						paragraph_text_area.replaceRange(
							lesson_text.substring(
							progress, progress + 1 ),
							cursor, cursor + 1 );
					}
				}
				else {
					// update the timings and statistics
					update_stats( char_wanted, false );

					// beep if required on wrong keys
					if ( option_mistake_beep.getState() )
						toolkit.beep();

					// update the paragraph typing area
					if ( char_wanted == '\n' ) {
						paragraph_text_area.insert(
							BADCR_INDICATOR, cursor );
						cursor++;
						// advance the cursors
						progress++; cursor++;
					}
					else {
						paragraph_text_area.replaceRange(
							BADKEY_INDICATOR,
							cursor, cursor + 1 );
						// advance the cursors
						progress++; cursor++;
					}

					// indicate the mistake on the keyboard
					// window
					if ( keyboard != null )
						keyboard.char_error(
								char_struck );
				}
			}

			// highlight progress through the exercise
			show_typing_cursor();
			show_expected_character();
		}
	}


	/*
	 * Reset selection in text areas when focus (re)gained;
	 * this gives a way to redisplay the expected next character.
	 */
	private void handle_text_got_focus() {
		show_typing_cursor();
		show_expected_character();

		// if we can, also pull focus into the relevant text entry
		// area for the lesson we have
		if ( lesson instanceof DrillLesson )
			entry_text_area.requestFocus();
		else if ( lesson instanceof ParagraphLesson )
			paragraph_text_area.requestFocus();
	}


	/*
	 * Update the label that gives feedback on what modifier
	 * keys are pressed.
	 */
	private void modifiers_update( KeyEvent evt, boolean pressed ) {

		// do nothing if the keypress doesn't actually indicate
		// that one of he modifier keys was pressed
		int keycode = evt.getKeyCode();
		if ( keycode == KeyEvent.VK_SHIFT
				|| keycode == KeyEvent.VK_CONTROL
				|| keycode == KeyEvent.VK_ALT
				|| keycode == KeyEvent.VK_CAPS_LOCK ) {

			// note the shift state before building the string
			boolean	shift		= evt.isShiftDown();
			boolean control		= evt.isControlDown();
			boolean alt		= evt.isAltDown();

			// ...now add in the one just pressed, or take out the
			// one that was just released
			if ( keycode == KeyEvent.VK_SHIFT )
				shift	= pressed;
			else if ( keycode == KeyEvent.VK_CONTROL )
				control	= pressed;
			else if ( keycode == KeyEvent.VK_ALT )
				alt	= pressed;

			// set the label to reflect the modifiers state
			modifiers_label.setText(
			( caps_lock ?
			KeyEvent.getKeyText( KeyEvent.VK_CAPS_LOCK ) + " " : "" )
			+ ( control ?
			KeyEvent.getKeyText( KeyEvent.VK_CONTROL ) + " " : "" )
			+ ( alt ?
			KeyEvent.getKeyText( KeyEvent.VK_ALT ) + " " : "" )
			+ ( shift ?
			KeyEvent.getKeyText( KeyEvent.VK_SHIFT ) + " " : "" ) );
		}
	}


	/*
	 * Handle GUI component actions and events.
	 */
	public void actionPerformed( ActionEvent evt ) {
		if ( evt.getSource() instanceof ImageButton ) {
			if ( evt.getSource() == pause_button ) {
				handle_menu_pause();
			}
			else if ( evt.getSource() == restart_button ) {
				handle_menu_restart();
			}
		}
		else if ( evt.getSource() instanceof MenuItem ) {
			if ( evt.getSource() == file_close ) {
				handle_menu_close();
			}
			else if ( evt.getSource() == file_pause ) {
				handle_menu_pause();
			}
			else if ( evt.getSource() == file_restart ) {
				handle_menu_restart();
			}
			else {
				Lesson lesson = (Lesson)(history_menuitems.get(
							evt.getSource() ));
				if ( lesson != null
					&& lesson instanceof ExerciseLesson ) {
					reveal( (ExerciseLesson)lesson );
				}
			}
		}
	}
	public void windowClosing( WindowEvent evt ) {
		handle_menu_close();
	}
	public void windowClosed( WindowEvent evt ) {}
	public void windowOpened( WindowEvent evt ) {
		handle_text_got_focus();
	}
	public void windowIconified( WindowEvent evt ) {}
	public void windowDeiconified( WindowEvent evt ) {
		handle_text_got_focus();
	}
	public void windowActivated( WindowEvent evt ) {
		handle_text_got_focus();
	}
	public void windowDeactivated( WindowEvent evt ) {}
	public void keyPressed( KeyEvent evt ) {

		// WORKROUND
		// In AWT 1.1, on UNIX/Motif anyway, there seem to be not
		// one but two Caps Lock problems:
		//  a) the char codes in a KeyEvent are always lowercase no
		//     matter what the caps lock state of the keyboard, and
		//  b) although the spec says setting caps lock should give
		//     a keyPressed event, and clearing it should give a
		//     keyReleased event, it doesn't work that way.
		// Here we try to work round these by toggling a local flag
		// if caps lock is pressed, then altering the key char later
		// on to be the correct case.  If an AWT fixes (b) but not
		// (a) this workround needs to be modified (although we try
		// to ameliorate the bad effects by dropping any Caps Lock
		// release event that follows a Caps Lock pressed event
		// within one second with no intervening other characters).
		// If it fixes (a) but not (b), the workround should have no
		// bad effects.  In any case, we'll let the user switch it
		// off and at least be no worse off than if it wasn't here.
		// [ Use System.out.println( evt ); here to see what you get!]
		if ( option_capslock_workround.getState() ) {
			caps_lock_was_last_press_event = false;
			if ( evt.getKeyCode() == KeyEvent.VK_CAPS_LOCK ) {
				caps_lock = ! caps_lock;
				caps_lock_was_last_press_event = true;
				if ( caps_lock )
					caps_lock_timestamp =
						System.currentTimeMillis();
			}
		}
		else
			caps_lock = false;
		// END WORKROUND

		// update the oh-so-clever shift key status label
		modifiers_update( evt, true );

		// if ctrl is pressed, pass on, since this may be an
		// accelerator
		if ( ! evt.isControlDown() ) {

			// ignore key events that aren't printable, e.g. Shift.
			char key_pressed	= evt.getKeyChar();
			int keycode		= evt.getKeyCode();
			if ( key_pressed != KeyEvent.CHAR_UNDEFINED
					// WORKROUND
					// For some reason, checking printabil-
					// ity isn't enough, since some JVMs
					// seem to consider Ctrl, Alt, and so
					// forth to be printable.  So, we also
					// check explicitly for at least the
					// most common modifiers
					&& keycode != KeyEvent.VK_SHIFT
					&& keycode != KeyEvent.VK_CONTROL
					&& keycode != KeyEvent.VK_ALT
					// END WORKROUND
				) {

				// make a translation across layouts if we have
				// a service to do that for us
				if ( translator != null ) {

					// derive a shift state from the event
					KeyShiftState shiftstate =
							KeyShiftState.UNSHIFTED;
					if ( evt.isAltDown() )
						shiftstate =
							KeyShiftState.META_ALT;
					else if ( evt.isShiftDown() )
						shiftstate =
							KeyShiftState.SHIFTED;

					// translate the char with this shift
					key_pressed = translator.translate(
								key_pressed,
								shiftstate );
				}

				// WORKROUND
				// Apply the caps lock workround to character
				if ( option_capslock_workround .getState() ) {
					if ( caps_lock || evt.isShiftDown() )
						key_pressed =
							Character.toUpperCase(
							key_pressed );
					else
						key_pressed =
							Character.toLowerCase(
							key_pressed );
				}
				// END WORKROUND

				// deal with the key press completely
				handle_key_press( key_pressed );

				// prevent the text areas (uneditable) from
				// trying to handle the key event too - this is
				// why we override keyPressed, and not the more
				// expected (and much easier) keyTyped
				evt.consume();
			}
		}
	}
	public void keyTyped( KeyEvent evt ) {}
	public void keyReleased( KeyEvent evt ) {

		// WORKROUND
		// Here, we apply the second, and usually probably redundant
		// part of the Caps Lock workround.  If problem (b) is fixed,
		// we'll get Caps Lock pressed on the first press (the one
		// that 'sets' the state in normal operation), and released
		// on the second press.  What we do to avoid problems with
		// our workround if KeyEvents still don't contain uppercased
		// characters is to release our caps lock state if we
		// have actually seen some intervening characters, or if
		// more than a second has elapsed.
		if ( option_capslock_workround.getState() ) {
			if ( evt.getKeyCode() == KeyEvent.VK_CAPS_LOCK )
				if ( ! caps_lock_was_last_press_event
					|| System.currentTimeMillis() >
						caps_lock_timestamp + 1000 )
					caps_lock = false;
		}
		// END WORKROUND

		// update the shift key status label
		modifiers_update( evt, false );
	}
	public void focusGained( FocusEvent evt ) {
		handle_text_got_focus();
	}
	public void focusLost( FocusEvent evt ) {}


	/*
	 * Register a stats GUI whose refresh we should call when
	 * a lesson is completed.
	 */
	public void register_statistics( StatisticsGUI statistics ) {
		this.statistics = statistics;
	}

	/*
	 * Register a keyboard GUI who we should use to indicate what
	 * keys are expected, and what was typed, including errors.
	 */
	public void register_keyboard( KeyboardGUI keyboard ) {
		this.keyboard = keyboard;
	}

	/*
	 * Register a translation service to convert between keyboard layouts.
	 */
	public void register_translator( KBTranslator translator ) {
		this.translator = translator;
	}


	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		ExerciseGUI exercise = new ExerciseGUI( false, 
			new KBStatistics(
				KBLayout.create_kblayout( "US" ) ));
		exercise.pack();
		exercise.reveal( new DrillLesson (
			"Test drill", "Test exercise character set",
			"The following tests the full character set:\n" +
			"`1234567890-=qwertyuiop[]\\asdfghjkl;'zxcvbnm,./\n" +
			"~!@#$%^&*()_+QWERTYUIOP{}|ASDFGHJKL:\"ZXCVBNM<>?\243\n"
		));
	}
}
